/*
 * Copyright (C) 2004   Derrick Moser
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include "Composer.h"

static Pattern part1Pattens[] =
{
  { "g", 0 },
  { "G", 1 },
  { "n", 2 },
  { "d", 3 },
  { "D", 4 },
  { "l", 5 },
  { "r", 5 },
  { "m", 6 },
  { "b", 7 },
  { "B", 8 },
  { "s", 9 },
  { "S", 10 },
  { "j", 12 },
  { "J", 13 },
  { "ch", 14 },
  { "k", 15 },
  { "t", 16 },
  { "p", 17 },
  { "h", 18 }
};

static Pattern part2Pattens[] =
{
  { "a", 0 },
  { "ae", 1 },
  { "ya", 2 },
  { "yae", 3 },
  { "eo", 4 },
  { "e", 5 },
  { "yeo", 6 },
  { "ye", 7 },
  { "o", 8 },
  { "wa", 9 },
  { "wae", 10 },
  { "oe", 11 },
  { "yo", 12 },
  { "u", 13 },
  { "wo", 14 },
  { "we", 15 },
  { "wi", 16 },
  { "yu", 17 },
  { "eu", 18 },
  { "ui", 19 },
  { "i", 20 }
};

static Pattern part3Pattens[] =
{
  { "g", 1 },
  { "G", 2 },
  { "gs", 3 },
  { "n", 4 },
  { "nj", 5 },
  { "nh", 6 },
  { "d", 7 },
  { "l", 8 },
  { "r", 8 },
  { "lg", 9 },
  { "rg", 9 },
  { "lm", 10 },
  { "rm", 10 },
  { "lb", 11 },
  { "rb", 11 },
  { "ls", 12 },
  { "rs", 12 },
  { "lt", 13 },
  { "rt", 13 },
  { "lp", 14 },
  { "rp", 14 },
  { "lh", 15 },
  { "rh", 15 },
  { "m", 16 },
  { "b", 17 },
  { "bs", 18 },
  { "s", 19 },
  { "S", 20 },
  { "ng", 21 },
  { "j", 22 },
  { "ch", 23 },
  { "k", 24 },
  { "t", 25 },
  { "p", 26 },
  { "h", 27 }
};

static Pattern partialPattens[] =
{
  { "g", 0x3131 },
  { "G", 0x3132 },
  { "gs", 0x3133 },
  { "n", 0x3134 },
  { "nj", 0x3135 },
  { "nh", 0x3136 },
  { "d", 0x3137 },
  { "D", 0x3138 },
  { "l", 0x3139 },
  { "r", 0x3139 },
  { "lg", 0x313a },
  { "rg", 0x313a },
  { "lm", 0x313b },
  { "rm", 0x313b },
  { "lb", 0x313c },
  { "rb", 0x313c },
  { "ls", 0x313d },
  { "rs", 0x313d },
  { "lt", 0x313e },
  { "rt", 0x313e },
  { "lp", 0x313f },
  { "rp", 0x313f },
  { "lh", 0x3140 },
  { "rh", 0x3140 },
  { "m", 0x3141 },
  { "b", 0x3142 },
  { "B", 0x3143 },
  { "bs", 0x3144 },
  { "s", 0x3145 },
  { "S", 0x3146 },
  { "ng", 0x3147 },
  { "j", 0x3148 },
  { "J", 0x3149 },
  { "ch", 0x314a },
  { "k", 0x314b },
  { "t", 0x314c },
  { "p", 0x314d },
  { "h", 0x314e },
  { "a", 0x314f },
  { "ae", 0x3150 },
  { "ya", 0x3151 },
  { "yae", 0x3152 },
  { "eo", 0x3153 },
  { "e", 0x3154 },
  { "yeo", 0x3155 },
  { "ye", 0x3156 },
  { "o", 0x3157 },
  { "wa", 0x3158 },
  { "wae", 0x3159 },
  { "oe", 0x315a },
  { "yo", 0x315b },
  { "u", 0x315c },
  { "wo", 0x315d },
  { "we", 0x315e },
  { "wi", 0x315f },
  { "yu", 0x3160 },
  { "eu", 0x3161 },
  { "ui", 0x3162 },
  { "i", 0x3163 }
};

SyllableParser::SyllableParser() :
  m_part1(sizeof(part1Pattens)/sizeof(Pattern), part1Pattens),
  m_part2(sizeof(part2Pattens)/sizeof(Pattern), part2Pattens),
  m_part3(sizeof(part3Pattens)/sizeof(Pattern), part3Pattens),
  m_partial(sizeof(partialPattens)/sizeof(Pattern), partialPattens)
{
}

SyllableParser::~SyllableParser()
{
}

static int compose(int r1, int r2, int r3)
{
  return 0xac00 + (r1 * 21 + r2) * 28 + r3;
}

int SyllableParser::parse(const int *s, int len, bool force)
{
  m_length = 0;
  m_result = -1;

  if(len > 0 && s[0] == '~')
  {
    int length = 1;
    int l = len - length;
    int s1 = m_partial.parse(&s[length], l, l, force);
    if(s1 != Success)
      return s1;
    length += m_partial.getLength();
    m_length = length;
    m_result = m_partial.getResult();
    return s1;
  }

  int s1 = m_part1.parse(s, len, len, force);
  if(s1 == Incomplete)
    return s1;

  int length = m_part1.getLength();
  int l = len - length;
  int s2 = m_part2.parse(&s[length], l, l, force);
  if(s2 != Success)
    return s2;

  length += m_part2.getLength();
  int r1 = m_part1.getResult();
  if(r1 < 0)
    r1 = 11;
  int r2 = m_part2.getResult();

  bool found_end = false;
  int i = 0;
  for(;;)
  {
    int start = length + i;
    l = len - start;
    int s12 = m_part1.parse(&s[start], l, l, false);
    if(s12 == Incomplete)
      break;

    if(s12 == Success)
    {
      int l12 = m_part1.getLength();
      int start2 = start + l12;
      if(start2 >= len)
        break;

      l = len - start2;
      int s22 = m_part2.parse(&s[start2], l, l, false);
      if(s22 != Failed)
      {
        found_end = true;
        break;
      }
    }
    ++i;
  }

  l = len - length;
  int l3 = ((found_end || !force) ? i : l);
  int s3 = m_part3.parse(&s[length], l, l3, found_end || force);

  if(s3 == Success)
  {
    length += m_part3.getLength();
    int r3 = m_part3.getResult();
    m_length = length;
    m_result = compose(r1, r2, r3);
  }
  else if(s3 == Failed)
  {
    m_length = length;
    m_result = compose(r1, r2, 0);
    return Success;
  }
  return s3;
}
